% MERLIN 2.0 Written by Henrik Bostrm in SICStus Prolog 3 #5 in 1997
% Deterministic version.

:- module(det_merlin2_module,[]).

:- use_module([misc_predicates,grammar_and_strings,intersection,grammar_to_program]).

/***** Main predicate ***************/

start(_,_,[],_,_,_,_,[]):-!.
start(ProgramFile,StartSymbol,Pos,_,BatchSize,Lambda,LookAhead,
      NewClauses):-
        init,
	asserta(start_symbol(StartSymbol)),
        asserta(batch_size(BatchSize)), 
        asserta(lambda(Lambda)), 
        asserta(look_ahead(LookAhead)), 
	grammar_and_strings(ProgramFile,Pos,[],CFG,PosStrings,_),
	induce_hmm(PosStrings,Automaton),
%        write('Merlin 2 Det. automaton:'),nl,write(Automaton),nl,
	start_state(StartState),
	grammar_intersection(StartSymbol,StartState,CFG,Automaton,NewCFG),
        grammar_to_program(StartSymbol,StartState,NewCFG,NewClauses), !.

init :-
    retractall(state_no(_)),
    retractall(start_symbol(_)),
    retractall(start_state(_)),
    retractall(state_name(_,_)),
    retractall(state(_)),
    retractall(symbol(_)),
    retractall(no_symbols(_)),
    retractall(no_states(_)),
    retractall(best_merge(_,_,_)),
    retractall(to_be_merged(_,_)),
    retractall(look_ahead(_)),
    retractall(initial(_)),
    retractall(state(_,_,_,_,_)),
    retractall(best(_)),
    retractall(temp(_)),
    retractall(batch_size(_)),
    retractall(lambda(_)), !.
    
/******************* HMM Induction **********************/

% In a first incremental phase, states may only be merged
% if they have identical emissions. In a second, non-incremental
% phase, all states may be merged.

induce_hmm(Strings,Automaton):-
    batch_size(BatchSize),
    lambda(Lambda),
    look_ahead(LookAhead),
    find_symbols(Strings),
    findall(S,symbol(S),Sigma),
    length(Sigma,NoSymbols),
    asserta(no_symbols(NoSymbols)),
    asserta(state_no(0)),
    asserta(no_states(0)),
    asserta(best(no_states(0))),
    find_hmm_incrementally(Strings,BatchSize,LookAhead,Lambda),
    find_best_hmm(LookAhead,0,Lambda,n),
    load_best_hmm,
    findall((Q,T,E),state(Q,_,_,T,E),HMM),
    findall(I,initial(I),Is),
    hmm_to_fsa(HMM,Is,Automaton), !.

find_symbols([]):-!.
find_symbols([String|Strings]):-
    check_string(String),
    find_symbols(Strings).

check_string([]):-!.
check_string([Symbol|String]):-
    update_symbols(Symbol), 
    check_string(String).    

update_symbols(Symbol):-
    symbol(Symbol), !.
update_symbols(Symbol):-
    asserta(symbol(Symbol)).

/******************* Find best HMM incrementally **************************/

find_hmm_incrementally([],_,_,_):-!.
find_hmm_incrementally(Strings,BatchSize,LookAhead,Lambda):-
    remove_batch(Strings,BatchSize,Batch,RemStrings),
%    write('***************** Next batch *********************'), nl,
    find_hmm(Batch,LookAhead,Lambda),
    find_hmm_incrementally(RemStrings,BatchSize,LookAhead,Lambda).

remove_batch([],_,[],[]):-!.
remove_batch(Strings,0,[],Strings):-!.
remove_batch([String|Strings],N1,[String|Batch],RemStrings):-
    N2 is N1-1,
    remove_batch(Strings,N2,Batch,RemStrings).

find_hmm([],_,_):-!.
find_hmm(Batch,LookAhead,Lambda):-
    build_initial_hmm(Batch),
    retract(best(no_states(_))),
    no_states(NoStates),
    asserta(best(no_states(NoStates))),
    retractall(original_probability(_)),
    log_probability(Lambda,OrigProb),
    asserta(original_probability(OrigProb)),
%    write('New initial HMM:'), nl,
%    show_states,
%    write('Original probability: '), write(OrigProb), nl,
    find_best_hmm(LookAhead,0,Lambda,y),
    load_best_hmm, !.

show_states :-
    write('------------------------'),nl,
    state(A,B,C,D,E), write(state(A,B,C,D,E)), nl, fail.
show_states :-
    write('Initial states: '), nl, initial(I), write(I), nl, fail.
show_states :-
    write('------------------------'),nl.

/******************* Initial HMM **************************/

build_initial_hmm([]):-!.
build_initial_hmm([String|Strings]):-
    include_string(String),
    build_initial_hmm(Strings).

include_string([Symbol|String]):-
    initial(Q1),
    get_state(Q1,_,_,_,E),
    member_chk((Symbol,_),E), !,
    update_state_emission(Q1,Symbol),
    include_string(Q1,String).
include_string(String):-
    new_state(Q1),
    asserta(initial(Q1)),
    asserta(best(initial(Q1))),
    build_hmm_branch(Q1,String).

include_string(Q1,[]):-
    !, update_state_transition(Q1,f).
include_string(Q1,[Symbol|String]):-
    get_state(Q1,_,_,T,_),
    successor(T,Symbol,Q2), !,
    update_state_transition(Q1,Q2),
    update_state_emission(Q2,Symbol),
    include_string(Q2,String).
include_string(Q1,String):-
    new_state(Q2),
    update_state_transition(Q1,Q2),
    build_hmm_branch(Q2,String).

successor(T,Symbol,Q):-
    member((Q,_),T),
    get_state(Q,_,_,_,E),
    member_chk((Symbol,_),E), !.

update_state_emission(Q,Symbol):-
    retract(state(Q,_,_,T,E)),
    retract(best(state(Q,_,_,T,E))),
    increase_counts(Symbol,E,NewE),
    length(T,NoT),
    length(NewE,NoNewE),
    beta_sigma(T,NoT,NewE,NoNewE,B,S),
    asserta(state(Q,B,S,T,NewE)),
    asserta(best(state(Q,B,S,T,NewE))), !.

increase_counts(S,[(S,N1)|R],[(S,N2)|R]):-
    !, N2 is N1+1.
increase_counts(S,[P|R],[P|NewR]):-
    increase_counts(S,R,NewR).

update_state_transition(Q1,Q2):-
    retract(state(Q1,_,_,T,E)),
    retract(best(state(Q1,_,_,T,E))),
    merge_counts(T,[(Q2,1)],NewT),
    length(NewT,NoNewT),
    length(E,NoE),
    beta_sigma(NewT,NoNewT,E,NoE,B,S),
    asserta(state(Q1,B,S,NewT,E)),
    asserta(best(state(Q1,B,S,NewT,E))), !.

new_state(Q1):-
    retract(state_no(Q0)),
    Q1 is Q0+1,
    asserta(state_no(Q1)), 
    retract(no_states(Q3)),
    Q4 is Q3+1,
    asserta(no_states(Q4)), !.

build_hmm_branch(Q,[Symbol]):-
    !, no_symbols(NoSymbols), 
    Sigma is -log(NoSymbols+1),
    asserta(state(Q,0,Sigma,[(f,1)],[(Symbol,1)])),
    asserta(best(state(Q,0,Sigma,[(f,1)],[(Symbol,1)]))).
build_hmm_branch(Q1,[Symbol|String]):-
    new_state(Q2),
    no_symbols(NoSymbols), 
    Sigma is -log(NoSymbols+1),
    asserta(state(Q1,0,Sigma,[(Q2,1)],[(Symbol,1)])),
    asserta(best(state(Q1,0,Sigma,[(Q2,1)],[(Symbol,1)]))),
    build_hmm_branch(Q2,String).

get_state(Q,B,S,T,E):-
    state(Q,B,S,T,E), !.

merge_counts([],T,T):-!.    
merge_counts(T,[],T):-!.    
merge_counts([(C,N1)|Cs1],[(C,N2)|Cs2],[(C,N3)|T]):-
    !, N3 is N1+N2,
    merge_counts(Cs1,Cs2,T).
merge_counts([(C1,N1)|Cs1],[(C2,N2)|Cs2],[(C1,N1)|T]):-
    C1 @< C2, !,
    merge_counts(Cs1,[(C2,N2)|Cs2],T).
merge_counts([(C1,N1)|Cs1],[(C2,N2)|Cs2],[(C2,N2)|T]):-
    merge_counts([(C1,N1)|Cs1],Cs2,T).

/********** Structure posterior probability *****************/

log_probability(Lambda,LogProbability):-
    no_states(NoStates),
    findall((-Lambda*NoT*log(NoStates+1)+B+Lambda*S),(state(_,B,S,T,_),length(T,NoT)),Terms),
    sum(Terms,0,LogProbability), !.

beta_sigma(T,NoTrans,E,NoEm,Beta,Sigma):-
    AlphaT is 1/NoTrans,
%    AlphaT is 1,
    beta_term(T,AlphaT,BetaTerm1),
    AlphaE is 1/NoEm,
%     AlphaE is 1,
    beta_term(E,AlphaE,BetaTerm2),
    no_symbols(NoSymbols),
    Beta is BetaTerm1+BetaTerm2,
    Sigma is -NoEm*log(NoSymbols+1), !.

beta_term(Cs,Alpha,BetaTerm):-
    log_gamma_sums(Cs,Alpha,0,LogGammaSum),
    log_gamma(1,LogGammaOne),
    sum_pairs(Cs,1,NoCs),
    log_gamma(NoCs,LogGammaCs),
    log_gamma(Alpha,LogGammaAlpha),
    length(Cs,N),
    BetaTerm is LogGammaSum+LogGammaOne-LogGammaCs-N*LogGammaAlpha.
    
log_gamma_sums([],_,Sum,Sum):-!.
log_gamma_sums([(_,C)|Cs],Alpha,Sum,LogGammaSum):-
    log_gamma(C+Alpha,LogGamma),
    NewSum is Sum+LogGamma,
    log_gamma_sums(Cs,Alpha,NewSum,LogGammaSum).

log_gamma(Z,LogGamma):-
    Sum is 0.99999999999980993227684700473478296744476168282198+
           676.52036812188509856700919044401903816411251975244084/Z
           -1259.13921672240287047156078755282840836424300664868028/(Z+1)+
           771.32342877765307884865282588943070775227268469602500/(Z+2)
           -176.61502916214059906584551353999392943274507608117860/(Z+3)+
           12.50734327868690481445893685327104972970563021816420/(Z+4)
           -0.13857109526572011689554706984971501358032683492780/(Z+5)+
           0.00000998436957801957085956266828104544089848531228/(Z+6)+
           0.00000015056327351493115583383579667028994545044040/(Z+7),
    LogGamma is log(Sum)+0.9189385332046727-Z-6.5+(Z-0.5)*log(Z+6.5).

sum_pairs([],S,S):-!.
sum_pairs([(_,N)|Ns],S1,S2):-
    S3 is S1+N,
    sum_pairs(Ns,S3,S2).

/******* HMM Merging ****************************/

find_best_hmm(0,_,_,_):-!.
find_best_hmm(LookAhead,BestGain,Lambda,Identical):-
    find_best_merge(Lambda,Gain,Identical), !,
    update_lookahead(Gain,BestGain,LookAhead,NewLookAhead),
    update_best_hmm(Gain,BestGain,NewBestGain),
    find_best_hmm(NewLookAhead,NewBestGain,Lambda,Identical).
find_best_hmm(_,_,_,_).

find_best_merge(Lambda,_,Identical):-
    original_probability(OrigProb),
    save_temp_hmm,
    asserta(best_merge(dummy,dummy,-inf)),
    findall(Q,state(Q,_,_,_,_),Qs),
    select_rest(Q1,Qs,Rest),
    get_state(Q1,_,_,_,E1),
    member(Q2,Rest),
    get_state(Q2,_,_,_,E2),
    check_emission(E1,E2,Identical),
    sort([Q1,Q2],[Q3,Q4]),
    load_temp_automaton_on_backtracking,
%    write('Testing: '), write((Q3,Q4)), nl,
    asserta(to_be_merged(Q3,Q4)),
    make_merges,
    log_probability(Lambda,NewProb),
    Gain is NewProb-OrigProb,
%    write('Gain: '), write(Gain), nl,
    best_merge(_,_,BestGainSoFar),
    Gain > BestGainSoFar,
    retract_one(best_merge(_,_,_)),
    asserta(best_merge(Q3,Q4,Gain)),
    fail.
find_best_merge(_,Gain,_):-
    retract(best_merge(Q1,Q2,Gain)), 
    Q1 \== dummy, !,
%    write('Best merge: '), write((Q1,Q2)), write(' Gain: '), write(Gain), nl,
    asserta(to_be_merged(Q1,Q2)),
    make_merges.

save_temp_hmm :-
    retractall(temp(_)),
    no_states(NoStates),
    asserta(temp(no_states(NoStates))),
    initial(Q),
    asserta(temp(initial(Q))),
    fail.
save_temp_hmm :-
    state(Q,B,S,T,E),
    asserta(temp(state(Q,B,S,T,E))),
    fail.
save_temp_hmm.

check_emission(_,_,n):-!.
check_emission(E1,E2,y):-
    identical_emission(E1,E2).

identical_emission([],[]):-!.
identical_emission([(S,_)|E1],[(S,_)|E2]):-
    identical_emission(E1,E2).

load_temp_automaton_on_backtracking.
load_temp_automaton_on_backtracking :-
    retractall(no_states(_)),
    retractall(initial(_)),
    retractall(state(_,_,_,_,_)),
    temp(Item),
    asserta(Item),
    fail.

make_merges :-
    retract(to_be_merged(Q1,Q2)), !,
%    write('Merging '), write((Q1,Q2)), nl,
    perform_merge(Q1,Q2),
    make_merges.
make_merges.

perform_merge(Q1,Q2):-
    retract_one(state(Q1,_,_,T1,E1)),
    retract_one(state(Q2,_,_,T2,E2)),
    retract_one(no_states(NoStates)),
    NewNoStates is NoStates-1,
    asserta(no_states(NewNoStates)),
    sort([Q1,Q2],[Q3,Q4]),
%    \+ forbidden_merge(Q3,Q4),
    merge_transitions(Q3,Q4,T1,T2,T3),
    merge_counts(E1,E2,E3),
    length(T3,NoT3),
    length(E3,NoE3),
    beta_sigma(T3,NoT3,E3,NoE3,B3,S3),
    asserta(state(Q3,B3,S3,T3,E3)),
%    update_forbidden_merges(Q3,Q4),
    update_transitions(Q3,Q4),
    update_initial_states(Q3,Q4), 
    update_to_be_merged(Q3,Q4), 
    check_nondeterminism(T3), !.

update_forbidden_merges(Q1,Q2):-
    retract(forbidden_merge(Q2,Q3)),
    \+ forbidden_merge(Q1,Q3),
    asserta(forbidden_merge(Q1,Q3)),
    fail.
update_forbidden_merges(Q1,Q2):-
    retract(forbidden_merge(Q3,Q2)),
    sort([Q1,Q3],[Q4,Q5]),
    \+ forbidden_merge(Q4,Q5),
    asserta(forbidden_merge(Q4,Q5)),
    fail.
update_forbidden_merges(_,_).

merge_transitions(Q1,Q2,T1,T2,T3):-
    merge_counts(T1,T2,Temp),
    new_transitions(Q1,Q2,Temp,T3), !.

new_transitions(Q1,Q2,T,NewT):-
    append(Before,[(Q1,N1)|Rest],T),
    append(Between,[(Q2,N2)|End],Rest), !,
    N3 is N1+N2,
    app(Between,End,NewRest),
    app(Before,[(Q1,N3)|NewRest],NewT).
new_transitions(Q1,Q2,T,NewT):-
    append(Before,[(Q2,N)|After],T), !,
    app([(Q1,N)|Before],After,TempT),
    sort(TempT,NewT).
new_transitions(_,_,T,T).

update_transitions(Q1,Q2):-
    retract(state(Q,B,S,T,E)),
    new_transitions(Q1,Q2,T,E,B,NewT,NewB),
    asserta(state(Q,NewB,S,NewT,E)),
    fail.
update_transitions(_,_).    

new_transitions(Q1,Q2,T,E,_,NewT,NewB):-
    append(Before,[(Q1,N1)|Rest],T),
    append(Between,[(Q2,N2)|End],Rest), !,
    N3 is N1+N2,
    app(Between,End,NewRest),
    app(Before,[(Q1,N3)|NewRest],NewT),
    length(NewT,NoNewT),
    length(E,NoE),
    beta_sigma(NewT,NoNewT,E,NoE,NewB,_).
new_transitions(Q1,Q2,T,E,_,NewT,NewB):-
    append(Before,[(Q2,N)|After],T), !,
    app([(Q1,N)|Before],After,TempT),
    sort(TempT,NewT),
    length(NewT,NoNewT),
    length(E,NoE),
    beta_sigma(NewT,NoNewT,E,NoE,NewB,_).
new_transitions(_,_,T,_,B,T,B).

update_initial_states(Q1,Q2):-
    initial(Q1), !,
    remove(initial(Q2)).
update_initial_states(Q1,Q2):-
    retract_one(initial(Q2)), !,
    asserta(initial(Q1)).
update_initial_states(_,_).

update_to_be_merged(Q1,Q2):-
    retract(to_be_merged(Q2,Q3)),
    \+ to_be_merged(Q1,Q3),
    asserta(to_be_merged(Q1,Q3)),
    fail.
update_to_be_merged(Q1,Q2):-
    retract(to_be_merged(Q3,Q2)),
    sort([Q1,Q3],[Q4,Q5]),
    \+ to_be_merged(Q4,Q5),
    asserta(to_be_merged(Q4,Q5)),
    fail.
update_to_be_merged(_,_).

check_nondeterminism(T):-
    select_rest((Q1,_),T,Rest),
    get_state(Q1,_,_,_,[(E,_)]),
    member((Q2,_),Rest),
    get_state(Q2,_,_,_,[(E,_)]),
    \+ to_be_merged(Q1,Q2),
    asserta(to_be_merged(Q1,Q2)),
    fail.
check_nondeterminism(_).

update_lookahead(Gain,BestGain,_,LookAhead):-
    Gain > BestGain, !,
    look_ahead(LookAhead).
update_lookahead(_,_,LookAhead,NewLookAhead):-
%    write('Looking ahead!'), nl,
    NewLookAhead is LookAhead-1.

update_best_hmm(Gain,BestGain,Gain):-
    Gain > BestGain, !,
%    write('Found better HMM!'), nl,
    update_best_hmm, !.
%    show_states.
update_best_hmm(_,BestGain,BestGain).

update_best_hmm :-
    retractall(best(_)),
    no_states(NoStates),
    asserta(best(no_states(NoStates))),
    initial(Q),
    asserta(best(initial(Q))),
    fail.
update_best_hmm :-
    state(Q,B,S,T,E),
    asserta(best(state(Q,B,S,T,E))),
    fail.
/*
update_best_hmm :-
    forbidden_merge(Q1,Q2),
    asserta(best(forbidden_merge(Q1,Q2))),
    fail.
*/
update_best_hmm.

load_best_hmm :-
    retractall(no_states(_)),
    retractall(initial(_)),
    retractall(state(_,_,_,_,_)),
%    retractall(forbidden_merge(_,_)),
    best(Item),
    asserta(Item),
    fail.
load_best_hmm.

/****************** HMM to FSA *********************/

hmm_to_fsa(HMM,Is,FSA):-
    asserta(start_state(0)),
    initial_transitions(Is,T0),
    update_final_states([(0,T0,[])|HMM],[(0,T0,[])|HMM],NewHMM),
    find_arcs(NewHMM,NewHMM,FSA), !.

initial_transitions([],[]):-!.
initial_transitions([I|Is],[(I,0)|T]):-
    initial_transitions(Is,T).

% Every state Q pointing to f should be transformed into final(Q)
% and all arcs should be updated!

update_final_states([],HMM,HMM):-!.
update_final_states([(Q,T,_)|QTEs],HMM1,HMM2):-
    member_chk((f,_),T), !,
    update_final_state(Q,HMM1,HMM3),
    update_final_states(QTEs,HMM3,HMM2).
update_final_states([_|QTEs],HMM1,HMM2):-
    update_final_states(QTEs,HMM1,HMM2).

update_final_state(_,[],[]):-!.
update_final_state(Q1,[(Q2,T2,E)|HMM1],[(NewQ,NewT,E)|HMM2]):-
    replace_final_state(Q1,Q2,NewQ),
    update_final_transitions(Q1,T2,NewT),
    update_final_state(Q1,HMM1,HMM2).

replace_final_state(Q,Q,final(Q)):-!.
replace_final_state(_,Q,Q).

update_final_transitions(_,[],[]):-!.
update_final_transitions(Q1,[(Q2,N)|T],[(NewQ,N)|NewT]):-
    replace_final_state(Q1,Q2,NewQ),    
    update_final_transitions(Q1,T,NewT).

find_arcs([],_,[]):-!.
find_arcs([(Q,T,_)|QTEs],HMM,FSA):-
    arcs_from(Q,T,HMM,ArcsFromQ),
    app(ArcsFromQ,RestFSA,FSA),
    find_arcs(QTEs,HMM,RestFSA).

arcs_from(_,[],_,[]):-!.
arcs_from(Q1,[(Q2,_)|T],HMM,ArcsFromQ1):-
    member_chk((Q2,_,E),HMM), !,
    arcs_from_to(Q1,Q2,E,ArcsFromQ1ToQ2),
    app(ArcsFromQ1ToQ2,RestArcs,ArcsFromQ1),
    arcs_from(Q1,T,HMM,RestArcs).
arcs_from(Q1,[_|T],HMM,ArcsFromQ1):-
    arcs_from(Q1,T,HMM,ArcsFromQ1).

arcs_from_to(_,_,[],[]):-!.
arcs_from_to(Q1,Q2,[(Symbol,_)|E],[(Q1,Symbol,Q2)|ArcsFromQ1ToQ2]):-
    arcs_from_to(Q1,Q2,E,ArcsFromQ1ToQ2).

/******** Extra misc. predicates ***************/

remove(G):-
    retract(G), !.
remove(_).

retract_one(G):-
    retract(G), !.

