/********** Grammar to program   ****************/

:- module(grammar_to_program,[grammar_to_program/4]).

:- use_module([misc_predicates,grammar_and_strings]).

grammar_to_program(StartSymbol,StartState,Grammar,NewClauses):-
        init,
	make_clauses(Grammar,(StartSymbol,StartState,empty),Clauses1),
	assert_temp_clauses(Clauses1),
	unfold_clauses(Clauses1,Clauses2),
	retractall(temp_clause(_,_)),
	StartSymbol = Functor/Arity,
	functor(Target,Functor,Arity),
	remove_dead_clauses(Clauses2,Target,Clauses2,NewClauses), !.

init :-
    retractall(temp_clause(_,_)),
    retractall(counter(_)), !.

make_clauses([],_,[]):-!.
make_clauses([(NT,[ClauseNo|RHS])|Rules],Start,[NewClause|Clauses]):-
	clause_(ClauseNo,OrigClause),
	make_clause(OrigClause,NT,RHS,Start,NewClause), 
	make_clauses(Rules,Start,Clauses), !.

make_clause(:-(Head,Body),NT,RHS,Start,:-(NewHead,NewBody)):-
	!, Head =.. [_|L],
	predicate_symbol(NT,Start,Predicate),
	NewHead =.. [Predicate|L],
	make_body(Body,RHS,Start,NewBody).
make_clause(Head,NT,[],Start,NewHead):-
	Head =.. [_|L],
	predicate_symbol(NT,Start,Predicate),
	NewHead =.. [Predicate|L], !.

make_body(Body,[],_,Body):-!.
make_body((Goal,Body),RHS,Start,(Goal,NewBody)):-
	silent_predicate(Goal), !,
	make_body(Body,RHS,Start,NewBody).
make_body((Goal,Body),[NT|RHS],Start,(NewGoal,NewBody)):-
	!,Goal =.. [_|L],
	predicate_symbol(NT,Start,Predicate),
	NewGoal =.. [Predicate|L],
	make_body(Body,RHS,Start,NewBody).
make_body(Goal,[NT],Start,NewGoal):-
	!,Goal =.. [_|L],
	predicate_symbol(NT,Start,Predicate),
	NewGoal =.. [Predicate|L].

predicate_symbol(((P/A),S,empty),((P/A),S,empty),P):-!.
predicate_symbol(((P/_),S,empty),_,NewSymbol):-
	!, name(P,L1),
	get_name(S,L2),
	app(L1,[95|L2],L3),
	name(NewSymbol,L3).
predicate_symbol(((P/_),S,R),_,NewSymbol):-
	name(P,L1),
	get_name(S,L2),
	get_name(R,L3),
	app(L1,[95|L2],L4),
	app(L4,[95|L3],L5),
	name(NewSymbol,L5).

get_name(final(S),L):-
	!, name(S,L).
get_name(S,L):-
	name(S,L).

assert_temp_clauses([]):- !.
assert_temp_clauses([:-(Head,Body)|Clauses]):-
	asserta(temp_clause(Head,Body)),
	assert_temp_clauses(Clauses).
assert_temp_clauses([Head|Clauses]):-
	asserta(temp_clause(Head,true)),
	assert_temp_clauses(Clauses).

unfold_clauses([],[]):- !.
unfold_clauses([:-(Head,Body)|Clauses],[NewClause|NewClauses]):-
	!, unfold_body(Body,NewBody),
	new_clause(Head,NewBody,NewClause),
	unfold_clauses(Clauses,NewClauses).
unfold_clauses([Clause|Clauses],[Clause|NewClauses]):-
	unfold_clauses(Clauses,NewClauses).

unfold_body(true,true):- !.
unfold_body((Goal,Body),NewBody):-
	single_resolvent(Goal,Body,Resolvent), !,
	unfold_body(Resolvent,NewBody).
unfold_body((Goal,Body),NewBody):-
	!, unfold_body(Body,RestBody),
	app_goal(Goal,RestBody,NewBody).
unfold_body(Goal,NewBody):-
	single_resolvent(Goal,true,Resolvent), !,
	unfold_body(Resolvent,NewBody).
unfold_body(Goal,Goal).

single_resolvent(Goal,OrigBody,Resolvent):-
	\+ two_applicable_clauses(Goal),
	temp_clause(Goal,Body),
	app_goal(Body,OrigBody,Resolvent), !.

two_applicable_clauses(Goal):-
	retractall(counter(_)),
	asserta(counter(0)),
	temp_clause(Goal,_),
	increase_counter,
	counter(N),
	N > 1, 
	retract(counter(_)), !.

increase_counter :-
	retract(counter(N1)),
	N2 is N1+1,
	asserta(counter(N2)), !.

app_goal(true,Goal,Goal):-!.
app_goal(Goal,true,Goal):-!.
app_goal((Goal1,Goal2),Goal3,(Goal1,Goal4)):-
	!, app_goal(Goal2,Goal3,Goal4).
app_goal(Goal1,Goal2,(Goal1,Goal2)).

new_clause(Head,true,Head):- !.
new_clause(Head,Body,:-(Head,Body)).

remove_dead_clauses([],_,_,[]):- !.
remove_dead_clauses([Clause|Clauses],Target,OrigClauses,NewClauses):-
	head(Clause,Head),
	\+ Target=Head,
	\+ (member(:-(_,OrigBody),OrigClauses),goal_member(Head,OrigBody)), !,
	remove_dead_clauses(Clauses,Target,OrigClauses,NewClauses).
remove_dead_clauses([Clause|Clauses],Target,OrigClauses,[Clause|NewClauses]):-
	remove_dead_clauses(Clauses,Target,OrigClauses,NewClauses).

head(:-(Head,_),Head):-!.
head(Head,Head).

goal_member(Goal,Goal):- !.
goal_member(Goal,(Goal,_)):- !.
goal_member(Goal1,(_,Goal2)):-
	goal_member(Goal1,Goal2).
