% Written by Henrik Bostrom in SICStus Prolog 3.1

:- module(merlin1_module,[]).

:- use_module([misc_predicates,grammar_and_strings,intersection,grammar_to_program]).

start(_,_,[],_,[]):-!.
start(ProgramFile,StartSymbol,Pos,Neg,NewClauses):-
	init,
	asserta(start_symbol(StartSymbol)),
	grammar_and_strings(ProgramFile,Pos,Neg,CFG,PosStrings,NegStrings),
        produce_automaton(PosStrings,NegStrings,Automaton),
%        write('Merlin 1 automaton:'),nl,write(Automaton),nl,
        start_state(StartState),
%        write('Start state: '),write(StartState),nl,
	grammar_intersection(StartSymbol,StartState,CFG,Automaton,NewCFG),
%	write('Grammar intersection: '), nl,
%	write(NewCFG), nl,
	grammar_to_program(StartSymbol,StartState,NewCFG,NewClauses), !.
%        write('Program:'), nl,
%        write(NewClauses), !.
	
init :-
	retractall(temp_state(_)),
	retractall(arc(_,_,_,_)),
	retractall(state(_)),
	retractall(state_no(_)),
	asserta(state_no(1)),  
	retractall(start_symbol(_)),
	retractall(start_state(_)),
	retractall(best_automaton(_,_,_)),
	retractall(mergeable(_,_)),
	retractall(state_name(_,_)), !.

erase_clauses :-
	retract(asserted(Ref)),
	erase(Ref),
	fail.
erase_clauses.

/****** Produce automaton *************************/

produce_automaton(PosStrings,NegStrings,NewAutomaton):-
	make_unreduced_automaton(PosStrings,pos),
	make_unreduced_automaton(NegStrings,neg),
	asserta(start_state(1)),
	assert_mergeables,
	reduce_automaton,
	retractall(arc(_,neg,_,_)),
	findall((NewState1,Clause,NewState2),
		(arc(State1,_,Clause,State2),
		 rename_final_state(State1,NewState1),
		 rename_final_state(State2,NewState2)),
		 NewAutomaton), !.
		
make_unreduced_automaton([],_):-!.
make_unreduced_automaton([Proof|Proofs],Class):-
	add_arcs(Proof,1,Class),
	make_unreduced_automaton(Proofs,Class), !.

add_arcs([Clause],State,Class):-
	new_final_state(Class,FinalState),
	asserta(arc(State,Class,Clause,FinalState)), !.
add_arcs([Clause|Clauses],State,Class):-
	arc(State,_,Clause,NewState), !,
	add_arcs(Clauses,NewState,Class).
add_arcs([Clause|Clauses],State,Class):-
	new_state(NewState),
	asserta(arc(State,Class,Clause,NewState)),
	add_new_arcs(Clauses,NewState,Class), !.

new_final_state(pos,final_pos(NewNo)):-
	!, retract(state_no(OldNo)),
	NewNo is OldNo+1, 
	asserta(state_no(NewNo)), 
	asserta(state(final_pos(NewNo))).
new_final_state(neg,final_neg(NewNo)):-
	!, retract(state_no(OldNo)),
	NewNo is OldNo+1, 
	asserta(state_no(NewNo)), 
	asserta(state(final_neg(NewNo))).
			
new_state(NewNo):-
	retract(state_no(OldNo)),
	NewNo is OldNo+1, 
	asserta(state_no(NewNo)), 
	asserta(state(NewNo)), !.

add_new_arcs([Clause],State,Class):-
	new_final_state(Class,FinalState), 
	asserta(arc(State,Class,Clause,FinalState)), !.
add_new_arcs([Clause|Clauses],State,Class):-
	new_state(NewState),
	asserta(arc(State,Class,Clause,NewState)),
	add_new_arcs(Clauses,NewState,Class).

assert_mergeables :-
	findall(State,state(State),States),
	select_rest(State1,[1|States],Rest),
	member(State2,Rest),
	check_mergeable(State1,State2),
	asserta(mergeable(State1,State2)),
	fail.
assert_mergeables.

check_mergeable(State1,State2):-
	arc(State1,pos,Clause,_),
	arc(State2,pos,Clause,_), !.

reduce_automaton :-
	make_best_merge,
	reduce_automaton, !.
reduce_automaton.

make_best_merge :-
	asserta(best_automaton(dummy,dummy,inf)),
	state_no(StateNo),
	findall(arc(A,B,C,D),arc(A,B,C,D),Arcs),
	mergeable(State1,State2),
	restore_old_automaton_upon_backtracking(StateNo,Arcs),
	merge_states(State1,State2),
	make_deterministic(State1,State2),
	findall(1,arc(_,pos,_,_),PosArcs),
	length(PosArcs,NoArcs),
	best_automaton(_,_,BestNoSoFar),
	NoArcs < BestNoSoFar,
	retract(best_automaton(_,_,_)),
	asserta(best_automaton(State1,State2,NoArcs)),
	fail.
make_best_merge :-
	retract(best_automaton(State1,State2,_)),
	State1 \== dummy,
	merge_states(State1,State2),
	make_deterministic,
	retractall(temp_state(_)),
	update_mergeables, !.

restore_old_automaton_upon_backtracking(_,_).	
restore_old_automaton_upon_backtracking(StateNo,Arcs):-
	retractall(temp_state(_)),
	retractall(arc(_,_,_,_)),
	retractall(merged(_,_,_)),
	assert_clauses(Arcs),
	retractall(state_no(_)),
	asserta(state_no(StateNo)),
	fail.

assert_clauses([]):-!.
assert_clauses([Clause|Clauses]):-
	asserta(Clause),
	assert_clauses(Clauses).

merge_states(final_pos(_),final_neg(_)):-
	!, fail.
merge_states(final_neg(_),final_pos(_)):-
	!, fail.
merge_states(State1,State2):-
	new_state(State1,State2,NewState),
	asserta(temp_state(NewState)),
	assert(merged(State1,State2,NewState)),
	update_temp_state(State1),
	update_temp_state(State2),
	findall(arc(NewFromState,Class,Sigma,NewState),
		(retract(arc(FromState,Class,Sigma,State1)),
		update_state(FromState,State1,State2,NewState,NewFromState)),
		Arcs,Rest1),
	findall(arc(NewFromState,Class,Sigma,NewState),
		(retract(arc(FromState,Class,Sigma,State2)),
		update_state(FromState,State1,State2,NewState,NewFromState)),
		Rest1,Rest2),
	findall(arc(NewState,Class,Sigma,ToState),
		retract(arc(State1,Class,Sigma,ToState)),Rest2,Rest3),
	findall(arc(NewState,Class,Sigma,ToState),
		retract(arc(State2,Class,Sigma,ToState)),Rest3,[]),
	assert_new_arcs(Arcs), !.

new_state(final_neg(_),_,NewState):-
	!, new_final_state(neg,NewState).
new_state(final_pos(_),_,NewState):-
	!, new_final_state(pos,NewState).
new_state(_,final_neg(_),NewState):-
	!, new_final_state(neg,NewState).
new_state(_,final_pos(_),NewState):-
	!, new_final_state(pos,NewState).
new_state(_,_,NewState):-
	!, new_state(NewState).

update_temp_state(State):-
	retract(temp_state(State)), !.
update_temp_state(_).	
	
update_state(State1,State1,_,NewState,NewState):-!.
update_state(State2,_,State2,NewState,NewState):-!.
update_state(State,_,_,_,State).

assert_new_arcs([]):-!.
assert_new_arcs([arc(State1,pos,Sigma,State2)|Arcs1]):-
	!, asserta(arc(State1,pos,Sigma,State2)),
	delete(arc(State1,_,Sigma,State2),Arcs1,Arcs2),
	assert_new_arcs(Arcs2).
assert_new_arcs([arc(State1,neg,Sigma,State2)|Arcs]):-
	\+ member(arc(State1,_,Sigma,State2),Arcs), !,
	asserta(arc(State1,neg,Sigma,State2)),
	assert_new_arcs(Arcs).
assert_new_arcs([_|Arcs]):-
	assert_new_arcs(Arcs).

make_deterministic(_,_):-
	make_deterministic, !.
make_deterministic(State1,State2):-
	retract(mergeable(State1,State2)), 
	!, fail.

make_deterministic :-
	temp_state(TempState),
	arc(TempState,_,Clause,State1),
	arc(TempState,_,Clause,State2),
	State1 \== State2, !,
	merge_states(State1,State2),
	make_deterministic.
make_deterministic.

update_mergeables :-
	retract(merged(State1,State2,NewState)), !,
	update_start_state(State1,State2,NewState),
	findall(mergeable(NewState,OtherState),
		(retract(mergeable(State1,OtherState)),
		OtherState\==State2),Mergeables,Rest1),
	findall(mergeable(NewState,OtherState),
		(retract(mergeable(State2,OtherState)),
		OtherState\==State1),Rest1,Rest2),
	findall(mergeable(NewState,OtherState),
		(retract(mergeable(OtherState,State1)),
		OtherState\==State2),Rest2,Rest3),
	findall(mergeable(NewState,OtherState),
		(retract(mergeable(OtherState,State2)),
		OtherState\==State1),Rest3,[]),
	assert_mergeables(Mergeables),
	update_mergeables.
update_mergeables.

update_start_state(State1,_,NewState):-
	retract(start_state(State1)), !,
	asserta(start_state(NewState)).
update_start_state(_,State2,NewState):-
	retract(start_state(State2)), !,
	asserta(start_state(NewState)).
update_start_state(_,_,_).

assert_mergeables([]):-!.
assert_mergeables([mergeable(State1,State2)|Mergeables]):-
	State1 \== State2,
	\+ member(mergeable(State1,State2),Mergeables), !,
	asserta(mergeable(State1,State2)),
	assert_mergeables(Mergeables).
assert_mergeables([_|Mergeables]):-
	assert_mergeables(Mergeables).

rename_final_state(final_pos(StateNo),final(StateNo)):-!.
rename_final_state(State,State).

	
