% MERLIN 2.0 Written by Henrik Bostrm in SICStus Prolog 3 #5 in 1997

:- module(merlin2_module,[]).

:- use_module([misc_predicates,grammar_and_strings,intersection,grammar_to_program]).

/***** Main predicate ***************/

start(_,_,[],_,_,_,_,[]):-!.
start(ProgramFile,StartSymbol,Pos,_,BatchSize,Lambda,LookAhead,NewClauses):-
        init,
	asserta(start_symbol(StartSymbol)),
        asserta(batch_size(BatchSize)), 
        asserta(lambda(Lambda)), 
        asserta(look_ahead(LookAhead)), 
	grammar_and_strings(ProgramFile,Pos,[],CFG,PosStrings,_),
        induce_hmm(PosStrings,Automaton),
%        write('Merlin 2 non-det. automaton:'),nl,write(Automaton),nl,
        start_state(StartState),
	grammar_intersection(StartSymbol,StartState,CFG,Automaton,NewCFG),
	grammar_to_program(StartSymbol,StartState,NewCFG,NewClauses), !.

init :-
    retractall(start_symbol(_)),
    retractall(start_state(_)),
    retractall(state_name(_,_)),
    retractall(symbol(_)),
    retractall(no_symbols(_)),
    retractall(no_states(_)),
    retractall(best_merge(_,_)),
    retractall(look_ahead(_)),
    retractall(initial(_)),
    retractall(mergeable(_,_,_,_,_,_,_,_)),
    retractall(state(_,_,_,_,_)),
    retractall(best(_)),
    retractall(batch_size(_)),
    retractall(lambda(_)), !.
    
/******************* HMM Induction **********************/

% In a first incremental phase, states may only be merged
% if they have identical emissions. In a second, non-incremental
% phase, all states may be merged.

induce_hmm(Strings,Automaton):-
    batch_size(BatchSize),
    lambda(Lambda),
    look_ahead(LookAhead),
%    length(Strings,NoStrings),
    find_symbols(Strings),
    findall(S,symbol(S),Sigma),
    length(Sigma,NoSymbols),
    asserta(no_symbols(NoSymbols)),
    asserta(state_no(0)),
    asserta(no_states(0)),
    asserta(best(no_states(0))),
    find_hmm_incrementally(Strings,BatchSize,LookAhead,Lambda),
    update_mergeables,
    find_best_hmm(LookAhead,0,Lambda,n),
    load_best_hmm,
    findall((Q,T,E),state(Q,_,_,T,E),HMM),
    findall(I,initial(I),Is),
%    write('Final HMM: '), nl,
%    show_states,
    hmm_to_fsa(HMM,Is,Automaton), !.

find_symbols([]):-!.
find_symbols([String|Strings]):-
    check_string(String),
    find_symbols(Strings).

check_string([]):-!.
check_string([Symbol|String]):-
    update_symbols(Symbol), 
    check_string(String).    

update_symbols(Symbol):-
    symbol(Symbol), !.
update_symbols(Symbol):-
    asserta(symbol(Symbol)).

/******************* Find best HMM incrementally **************************/

find_hmm_incrementally([],_,_,_):-!.
find_hmm_incrementally(Strings,BatchSize,LookAhead,Lambda):-
    remove_batch(Strings,BatchSize,Batch,RemStrings),
%    remove_accepted_strings(Batch,NonAccepted),
    find_hmm(Batch,LookAhead,Lambda),
    find_hmm_incrementally(RemStrings,BatchSize,LookAhead,Lambda).

remove_batch([],_,[],[]):-!.
remove_batch(Strings,0,[],Strings):-!.
remove_batch([String|Strings],N1,[String|Batch],RemStrings):-
    N2 is N1-1,
    remove_batch(Strings,N2,Batch,RemStrings).

/*
remove_accepted_strings([],[]):-!.
remove_accepted_strings([String|Batch],[String|NonAccepted]):-
    \+ (initial(I), accepted(String,I)), !,
    remove_accepted_strings(Batch,NonAccepted).
remove_accepted_strings([_|Batch],NonAccepted):-
    remove_accepted_strings(Batch,NonAccepted).

accepted(String,Q):-
    cached(Q,String,Accepted), !,
    Accepted=y.
accepted([],Q):-
    !, Q=f.
accepted([Sigma|String],Q):-
    get_state(Q,_,_,T,E),
    member_chk((Sigma,_),E),
    member((NewQ,_),T),
    accepted(String,NewQ), 
    asserta(cached(Q,[Sigma|String],y)), !.
accepted(String,Q):-
    asserta(cached(Q,String,n)), !,
    fail.
*/

find_hmm([],_,_):-!.
find_hmm(NonAccepted,LookAhead,Lambda):-
    build_initial_hmm(NonAccepted,NewStates),
%    write('Initial HMM:'), nl, show_states,
%    length(NewStates,NoNewStates),
%    write('Added '), write(NoNewStates), write(' new states.'), nl,
    update_initial_mergeables(NewStates),
    retract(best(no_states(_))),
    no_states(NoStates),
    asserta(best(no_states(NoStates))),
    find_best_hmm(LookAhead,0,Lambda,y),
    load_best_hmm, !.

show_states :-
    state(Q,B,S,T,E),
    write(state(Q,B,S,T,E)), nl,
    fail.
show_states.

/******************* Initial HMM **************************/

build_initial_hmm([],[]):-!.
build_initial_hmm([String|Strings],[Q|NewStates]):-
    new_state(Q),
    asserta(initial(Q)),
    asserta(best(initial(Q))),
    build_hmm_branch(Q,String,FirstStates),
    app(FirstStates,RestStates,NewStates),
    build_initial_hmm(Strings,RestStates).

new_state(Q1):-
    retract(state_no(Q0)),
    Q1 is Q0+1,
    asserta(state_no(Q1)), 
    retract(no_states(Q3)),
    Q4 is Q3+1,
    asserta(no_states(Q4)), !.

build_hmm_branch(Q,[Symbol],[]):-
    !, no_symbols(NoSymbols), 
    Sigma is -log(NoSymbols+1),
    asserta(state(Q,0,Sigma,[(f,1)],[(Symbol,1)])),
    asserta(best(state(Q,0,Sigma,[(f,1)],[(Symbol,1)]))).
build_hmm_branch(Q1,[Symbol|String],[Q2|NewStates]):-
    new_state(Q2),
    no_symbols(NoSymbols), 
    Sigma is -log(NoSymbols+1),
    asserta(state(Q1,0,Sigma,[(Q2,1)],[(Symbol,1)])),
    asserta(best(state(Q1,0,Sigma,[(Q2,1)],[(Symbol,1)]))),
    build_hmm_branch(Q2,String,NewStates).

update_initial_mergeables(NewStates):- 
    findall(OldState,(state(OldState,_,_,_,_), \+ member(OldState,NewStates)),OldStates),
    select_state_pair(Q1,Q2,NewStates,OldStates),
    get_state(Q1,_,_,T1,E1), % No state points at Q1 and Q2
    get_state(Q2,_,_,T2,E2),
    identical_emission(E1,E2),    
    sort([Q1,Q2],[Q3,Q4]),
    merge_transitions(Q3,Q4,T1,T2,T3),
    merge_counts(E1,E2,E3),
    length(T3,NoT3),
    length(E3,NoE3),
    beta_sigma(T3,NoT3,E3,NoE3,B3,S3),
    length(T1,NoT1),
    length(T2,NoT2),
    asserta(mergeable(Q3,Q4,B3,S3,T3,E3,(NoStates,
			    (NoT1*log(NoStates+1)+NoT2*log(NoStates+1)-NoT3*
			     log(NoStates))),[])),
    asserta(best(mergeable(Q3,Q4,B3,S3,T3,E3,(NoStates,
			    (NoT1*log(NoStates+1)+NoT2*log(NoStates+1)-NoT3*
			     log(NoStates))),[]))),
    fail.
update_initial_mergeables(_).

select_state_pair(Q1,Q2,NewStates,OldStates):-
    select_rest(Q1,NewStates,Rest),
    member_in_either(Q2,Rest,OldStates).

member_in_either(E,L1,_):-
    member(E,L1).
member_in_either(E,_,L2):-
    member(E,L2).

get_state(Q,B,S,T,E):-
    state(Q,B,S,T,E), !.

identical_emission([],[]):-!.
identical_emission([(S,_)|E1],[(S,_)|E2]):-
    identical_emission(E1,E2).

merge_counts([],T,T):-!.    
merge_counts(T,[],T):-!.    
merge_counts([(C,N1)|Cs1],[(C,N2)|Cs2],[(C,N3)|T]):-
    !, N3 is N1+N2,
    merge_counts(Cs1,Cs2,T).
merge_counts([(C1,N1)|Cs1],[(C2,N2)|Cs2],[(C1,N1)|T]):-
    C1 @< C2, !,
    merge_counts(Cs1,[(C2,N2)|Cs2],T).
merge_counts([(C1,N1)|Cs1],[(C2,N2)|Cs2],[(C2,N2)|T]):-
    merge_counts([(C1,N1)|Cs1],Cs2,T).

/********** Structure posterior probability *****************/

beta_sigma(T,NoTrans,E,NoEm,Beta,Sigma):-
    AlphaT is 1/NoTrans,
%    AlphaT is 1,
    beta_term(T,AlphaT,BetaTerm1),
    AlphaE is 1/NoEm,
%     AlphaE is 1,
    beta_term(E,AlphaE,BetaTerm2),
    no_symbols(NoSymbols),
    Beta is BetaTerm1+BetaTerm2,
    Sigma is -NoEm*log(NoSymbols+1), !.

beta_term(Cs,Alpha,BetaTerm):-
    log_gamma_sums(Cs,Alpha,0,LogGammaSum),
    log_gamma(1,LogGammaOne),
    sum_pairs(Cs,1,NoCs),
    log_gamma(NoCs,LogGammaCs),
    log_gamma(Alpha,LogGammaAlpha),
    length(Cs,N),
    BetaTerm is LogGammaSum+LogGammaOne-LogGammaCs-N*LogGammaAlpha.
    
log_gamma_sums([],_,Sum,Sum):-!.
log_gamma_sums([(_,C)|Cs],Alpha,Sum,LogGammaSum):-
    log_gamma(C+Alpha,LogGamma),
    NewSum is Sum+LogGamma,
    log_gamma_sums(Cs,Alpha,NewSum,LogGammaSum).

log_gamma(Z,LogGamma):-
    Sum is 0.99999999999980993227684700473478296744476168282198+
           676.52036812188509856700919044401903816411251975244084/Z
           -1259.13921672240287047156078755282840836424300664868028/(Z+1)+
           771.32342877765307884865282588943070775227268469602500/(Z+2)
           -176.61502916214059906584551353999392943274507608117860/(Z+3)+
           12.50734327868690481445893685327104972970563021816420/(Z+4)
           -0.13857109526572011689554706984971501358032683492780/(Z+5)+
           0.00000998436957801957085956266828104544089848531228/(Z+6)+
           0.00000015056327351493115583383579667028994545044040/(Z+7),
    LogGamma is log(Sum)+0.9189385332046727-Z-6.5+(Z-0.5)*log(Z+6.5).

sum_pairs([],S,S):-!.
sum_pairs([(_,N)|Ns],S1,S2):-
    S3 is S1+N,
    sum_pairs(Ns,S3,S2).

/******* HMM Merging ****************************/

find_best_hmm(0,_,_,_):-!.
find_best_hmm(LookAhead,GainSoFar,Lambda,Identical):-
    find_best_merge(Lambda,Gain,Identical), !,
    update_lookahead(Gain,LookAhead,NewLookAhead),
    TotalGain is GainSoFar+Gain,
    update_best_hmm(Gain,TotalGain),
    find_best_hmm(NewLookAhead,TotalGain,Lambda,Identical).
find_best_hmm(_,_,_,_).

find_best_merge(Lambda,_,_):-
    asserta(best_merge(dummy,-inf)),
    no_states(NoStates),
    mergeable(Q1,Q2,B3,S3,T3,E3,(NoStates,Expression),Affected),
    affected_gain(Affected,NoStates,Lambda,0,AffectedGain),
    get_state(Q1,B1,S1,_,_),
    get_state(Q2,B2,S2,_,_),
    Gain is Lambda*Expression+Lambda*(S3-S1-S2)-B1-B2+B3+AffectedGain,
    best_merge(_,BestGainSoFar),
%    write('Testing: '), write((Q1,Q2)), write(' Gain: '), write(Gain), nl,
    Gain > BestGainSoFar,
%    write('Best merge so far: '), write((Q1,Q2)), write(' Gain: '),write(Gain), nl,
    retract_one(best_merge(_,_)),
    asserta(best_merge((Q1,Q2,B3,S3,T3,E3,Affected),Gain)),
    fail.
find_best_merge(_,Gain,Identical):-
    retract(best_merge(BestMerge,Gain)), 
    BestMerge \== dummy, !,
    perform_merge(BestMerge,Identical).

update_lookahead(Gain,_,LookAhead):-
    Gain > 0, !,
    look_ahead(LookAhead).
update_lookahead(_,LookAhead,NewLookAhead):-
%    write('Looking ahead!'), nl,
    NewLookAhead is LookAhead-1.

update_best_hmm(Gain,TotalGain):-
    Gain > 0, 
    TotalGain > 0, !,
    update_best_hmm.
update_best_hmm(_,_).

update_best_hmm :-
%    write('New best HMM:'), nl,
    retractall(best(_)),
    no_states(NoStates),
    asserta(best(no_states(NoStates))),
    initial(Q),
    asserta(best(initial(Q))),
    fail.
update_best_hmm :-
    state(Q,B,S,T,E),
%    write(state(Q,B,S,T,E)), nl,
    asserta(best(state(Q,B,S,T,E))),
    fail.
update_best_hmm :-
    mergeable(A,B,C,D,E,F,G,H),
    asserta(best(mergeable(A,B,C,D,E,F,G,H))),
    fail.
update_best_hmm.

affected_gain([],_,_,Gain,Gain):-!.
affected_gain([(_,OldB,NewB,_,(NoStates,Expression))|Affected],NoStates,Lambda,
	      GainSoFar,Gain):-
    NewGain is GainSoFar+Lambda*Expression-OldB+NewB,
    affected_gain(Affected,NoStates,Lambda,NewGain,Gain).

perform_merge((Q1,Q2,B3,S3,T3,E3,Affected),Identical):-
%    write('Merging: '),write((Q1,Q2)), nl,
    retract_one(state(Q1,_,_,T1,_)),
    retract_one(state(Q2,_,_,T2,_)),
    retract(no_states(NoStates)), !,
    NewNoStates is NoStates-1,
    asserta(no_states(NewNoStates)),
    retractall(mergeable(Q1,_,_,_,_,_,_,_)),
    retractall(mergeable(_,Q1,_,_,_,_,_,_)),
    retractall(mergeable(Q2,_,_,_,_,_,_,_)),
    retractall(mergeable(_,Q2,_,_,_,_,_,_)),
    asserta(state(Q1,B3,S3,T3,E3)),
    update_affected_states(Affected),
    update_affected_mergeables(Q1,Q2),
    remove_affected(Q1,T1),
    remove_affected(Q2,T2),
    add_affected(Q1,B3,T3,E3),
    update_transitions(Q1,Q2),
    new_mergeables(Q1,T3,E3,Identical),
    update_initial_states(Q1,Q2), !.

update_affected_states([]):-!.
update_affected_states([(Q,_,NewB,NewT,_)|Qs]):-
    retract_one(state(Q,_,S,_,E)),
    asserta(state(Q,NewB,S,NewT,E)),
    update_affected_states(Qs).

update_affected_mergeables(Q1,Q2):-
    retract(mergeable(Q3,Q4,B,S,T,E,Expression,Affected)),
    assert_new_mergeable(Q1,Q2,Q3,Q4,B,S,T,E,Expression,Affected),
    fail.
update_affected_mergeables(_,_).

assert_new_mergeable(Q1,Q2,Q3,Q4,_,S,T,E,_,Affected):-
    append(Before,[(Q1,N1)|Rest],T),
    append(Between,[(Q2,N2)|End],Rest), !,
    N3 is N1+N2,
    app(Between,End,NewRest),
    app(Before,[(Q1,N3)|NewRest],NewT),
    length(NewT,NoNewT),
    length(E,NoE),
    beta_sigma(NewT,NoNewT,E,NoE,NewB,_),
    get_state(Q3,_,_,T3,_),
    get_state(Q4,_,_,T4,_),
    length(T3,NoT3),
    length(T4,NoT4),
    asserta(mergeable(Q3,Q4,NewB,S,NewT,E,(NoStates,
			    (NoT3*log(NoStates+1)+NoT4*log(NoStates+1)-NoNewT*
			     log(NoStates))),Affected)).

assert_new_mergeable(Q1,Q2,Q3,Q4,B,S,T,E,Expression,Affected):-
    append(Before,[(Q2,N)|After],T), !,
    app([(Q1,N)|Before],After,TempT),
    sort(TempT,NewT),
    asserta(mergeable(Q3,Q4,B,S,NewT,E,Expression,Affected)).
assert_new_mergeable(_,_,Q3,Q4,B,S,T,E,Expression,Affected):-
    asserta(mergeable(Q3,Q4,B,S,T,E,Expression,Affected)).

merge_transitions(Q1,Q2,T1,T2,T3):-
    merge_counts(T1,T2,Temp),
    new_transitions(Q1,Q2,Temp,T3), !.

new_transitions(Q1,Q2,T,NewT):-
    append(Before,[(Q1,N1)|Rest],T),
    append(Between,[(Q2,N2)|End],Rest), !,
    N3 is N1+N2,
    app(Between,End,NewRest),
    app(Before,[(Q1,N3)|NewRest],NewT).
new_transitions(Q1,Q2,T,NewT):-
    append(Before,[(Q2,N)|After],T), !,
    app([(Q1,N)|Before],After,TempT),
    sort(TempT,NewT).
new_transitions(_,_,T,T).

update_transitions(Q1,Q2):-
    retract(state(Q,B,S,T,E)),
    replace_transitions(Q2,Q1,T,NewT),
    asserta(state(Q,B,S,NewT,E)),
    fail.
update_transitions(_,_).    

replace_transitions(_,_,[],[]):-!.
replace_transitions(Q2,Q1,[(Q2,N)|T],[(Q1,N)|T]):-!.
replace_transitions(Q2,Q1,[Q3|T1],[Q3|T2]):-
    replace_transitions(Q2,Q1,T1,T2).

remove_affected(Q1,T):-
    select_rest((Q2,_),T,Rest),
    member((Q3,_),Rest),
    retract_one(mergeable(Q2,Q3,B3,S3,T3,E3,Expression,OldAffected)),
    delete_chk((Q1,_,_,_,_),OldAffected,NewAffected),
    asserta(mergeable(Q2,Q3,B3,S3,T3,E3,Expression,NewAffected)),
    fail.
remove_affected(_,_).

add_affected(Q1,B,T,E):-
    select_rest((Q2,_),T,Rest),
    member((Q3,_),Rest),
    retract_one(mergeable(Q2,Q3,B3,S3,T3,E3,Expression,Affected)),
    new_transitions(Q2,Q3,T,NewT), 
    length(NewT,NoNewT), 
    length(E,NoE), 
    beta_sigma(NewT,NoNewT,E,NoE,NewB,_),   
    asserta(mergeable(Q2,Q3,B3,S3,T3,E3,Expression,
		     [(Q1,B,NewB,NewT,(NoStates,
		       ((NoNewT+1)*log(NoStates+1)-NoNewT*log(NoStates+1))))|Affected])),
    fail.
add_affected(_,_,_,_).

new_mergeables(Q1,T1,E1,Identical):- % Some state may point at Q1 and Q2
    state(Q2,_,_,T2,E2),
    Q1 \== Q2,
    check_emission(E1,E2,Identical),     
    sort([Q1,Q2],[Q3,Q4]),
    merge_transitions(Q3,Q4,T1,T2,T3),
    merge_counts(E1,E2,E3),
    length(T3,NoT3),
    length(E3,NoE3),
    beta_sigma(T3,NoT3,E3,NoE3,B3,S3),
    length(T1,NoT1),
    length(T2,NoT2),
    find_affected(Q3,Q4,Affected),
    asserta(mergeable(Q3,Q4,B3,S3,T3,E3,(NoStates,
			    (NoT1*log(NoStates+1)+NoT2*log(NoStates+1)-NoT3*
			     log(NoStates))),Affected)),
    fail.
new_mergeables(_,_,_,_).

check_emission(_,_,n):-!.
check_emission(E1,E2,y):-
    identical_emission(E1,E2).

find_affected(Q1,Q2,Affected):-
    findall((Q,B,NewB,NewT,(NoStates,
			    ((NoNewT+1)*log(NoStates+1)-NoNewT*log(NoStates+1)))), 
	    (state(Q,B,_,T,E), Q \== Q1, Q \== Q2, check_transitions(Q1,Q2,T,NewT), 
	     length(NewT,NoNewT), length(E,NoE), 
	     beta_sigma(NewT,NoNewT,E,NoE,NewB,_)), 
	    Affected), !.

check_transitions(Q1,Q2,T,NewT):-
	   append(Before,[(Q1,N1)|Rest],T),
	   append(Between,[(Q2,N2)|End],Rest), !,
	   N3 is N1+N2,
	   app(Between,End,NewRest),
	   app(Before,[(Q1,N3)|NewRest],NewT).

update_initial_states(Q1,Q2):-
    initial(Q1), !,
    remove(initial(Q2)).
update_initial_states(Q1,Q2):-
    retract_one(initial(Q2)), !,
    asserta(initial(Q1)).
update_initial_states(_,_).

load_best_hmm :-
    retractall(no_states(_)),
    retractall(initial(_)),
    retractall(state(_,_,_,_,_)),
    retractall(mergeable(_,_,_,_,_,_,_,_)),
    best(Item),
    asserta(Item),
    fail.
load_best_hmm.

update_mergeables :-
    findall(Q,state(Q,_,_,_,_),Qs),
    select_rest(Q1,Qs,Rest),
    get_state(Q1,_,_,T1,E1),
    member(Q2,Rest),
    get_state(Q2,_,_,T2,E2),
    \+ identical_emission(E1,E2),
    sort([Q1,Q2],[Q3,Q4]),
    merge_transitions(Q3,Q4,T1,T2,T3),
    merge_counts(E1,E2,E3),
    length(T3,NoT3),
    length(E3,NoE3),
    beta_sigma(T3,NoT3,E3,NoE3,B3,S3),
    length(T1,NoT1),
    length(T2,NoT2),
    find_affected(Q3,Q4,Affected),
    asserta(mergeable(Q3,Q4,B3,S3,T3,E3,(NoStates,
			    (NoT1*log(NoStates+1)+NoT2*log(NoStates+1)-NoT3*
			     log(NoStates))),Affected)),
    fail.    
update_mergeables.

/****************** HMM to FSA *********************/

hmm_to_fsa(HMM,Is,FSA):-
    asserta(start_state(0)),
    initial_transitions(Is,T0),
    update_final_states([(0,T0,[])|HMM],[(0,T0,[])|HMM],NewHMM),
    find_arcs(NewHMM,NewHMM,FSA), !.

initial_transitions([],[]):-!.
initial_transitions([I|Is],[(I,0)|T]):-
    initial_transitions(Is,T).

% Every state Q pointing to f should be transformed into final(Q)
% and all arcs should be updated!

update_final_states([],HMM,HMM):-!.
update_final_states([(Q,T,_)|QTEs],HMM1,HMM2):-
    member_chk((f,_),T), !,
    update_final_state(Q,HMM1,HMM3),
    update_final_states(QTEs,HMM3,HMM2).
update_final_states([_|QTEs],HMM1,HMM2):-
    update_final_states(QTEs,HMM1,HMM2).

update_final_state(_,[],[]):-!.
update_final_state(Q1,[(Q2,T2,E)|HMM1],[(NewQ,NewT,E)|HMM2]):-
    replace_final_state(Q1,Q2,NewQ),
    update_final_transitions(Q1,T2,NewT),
    update_final_state(Q1,HMM1,HMM2).

replace_final_state(Q,Q,final(Q)):-!.
replace_final_state(_,Q,Q).

update_final_transitions(_,[],[]):-!.
update_final_transitions(Q1,[(Q2,N)|T],[(NewQ,N)|NewT]):-
    replace_final_state(Q1,Q2,NewQ),    
    update_final_transitions(Q1,T,NewT).

find_arcs([],_,[]):-!.
find_arcs([(Q,T,_)|QTEs],HMM,FSA):-
    arcs_from(Q,T,HMM,ArcsFromQ),
    app(ArcsFromQ,RestFSA,FSA),
    find_arcs(QTEs,HMM,RestFSA).

arcs_from(_,[],_,[]):-!.
arcs_from(Q1,[(Q2,_)|T],HMM,ArcsFromQ1):-
    member_chk((Q2,_,E),HMM), !,
    arcs_from_to(Q1,Q2,E,ArcsFromQ1ToQ2),
    app(ArcsFromQ1ToQ2,RestArcs,ArcsFromQ1),
    arcs_from(Q1,T,HMM,RestArcs).
arcs_from(Q1,[_|T],HMM,ArcsFromQ1):-
    arcs_from(Q1,T,HMM,ArcsFromQ1).

arcs_from_to(_,_,[],[]):-!.
arcs_from_to(Q1,Q2,[(Symbol,_)|E],[(Q1,Symbol,Q2)|ArcsFromQ1ToQ2]):-
    arcs_from_to(Q1,Q2,E,ArcsFromQ1ToQ2).

/**** Extra misc. predicates ******/

remove(G):-
    retract(G), !.
remove(_).

retract_one(G):-
    retract(G), !.

